package de.hft.stuttgart.citydoctor2.datastructure;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.utils.CityGmlUtils;
import de.hft.stuttgart.citydoctor2.utils.CopyHandler;
import de.hft.stuttgart.citydoctor2.utils.Copyable;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.citygml4j.core.model.core.AbstractSpaceBoundaryProperty;
import org.citygml4j.core.util.geometry.GeometryFactory;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurface;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurfaceProperty;
import org.xmlobjects.gml.model.geometry.primitives.Solid;
import org.xmlobjects.gml.model.geometry.primitives.SolidProperty;

import java.io.Serial;
import java.util.ArrayList;
import java.util.List;

public class TunnelConstructiveElement extends CityObject {

    private static final Logger logger = LogManager.getLogger(TunnelConstructiveElement.class);

    private static final String CANNOT_ADD = "Cannot add ";

    @Serial
    private static final long serialVersionUID = 7353233899458901155L;

    private final org.citygml4j.core.model.tunnel.TunnelConstructiveElement gmlTunnelElement;
    private final List<BoundarySurface> boundarySurfaceList = new ArrayList<>();
    private AbstractTunnel parent;

    public TunnelConstructiveElement(org.citygml4j.core.model.tunnel.TunnelConstructiveElement gmlObject) {
        this.gmlTunnelElement = gmlObject;
    }

    public void setParent(AbstractTunnel parent) {
        this.parent = parent;
    }

    public AbstractTunnel getParent() {
        return parent;
    }

    @Override
    public Copyable createCopyInstance() {
        return new TunnelConstructiveElement(gmlTunnelElement);
    }

    @Override
    public void reCreateGeometries(GeometryFactory factory, ParserConfiguration config) {
        // only handles CityGML2 for now
        // unknown which CityGML is handled here
        // need context information to decide
        for (Geometry geom : getGeometries()) {
            if (geom instanceof ImplicitGeometryHolder) {
                continue;
            }
            switch (geom.getType()) {
                case SOLID:
                    Solid solid = CityGmlUtils.createSolid(geom, factory, config);
                    setSolidAccordingToLod(geom, solid);
                    break;
                case MULTI_SURFACE:
                    MultiSurface ms = CityGmlUtils.createMultiSurface(geom, factory, config);
                    setMultiSurfaceAccordingToLod(geom, ms);
                    break;
                case COMPOSITE_SURFACE:
                	throw new IllegalStateException("Tunnel constructive element cannot have a composite surface geometry");
            }
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            reCreateBoundarySurface(factory, config, bs);
        }
    }

    @Override
    public void accept(Check c) {
        super.accept(c);
        if (c.canExecute(this)) {
            c.check(this);
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.accept(c);
        }
    }

    @Override
    public void collectContainedErrors(List<CheckError> errors) {
        super.collectContainedErrors(errors);
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.collectContainedErrors(errors);
        }
    }

    @Override
    public void clearAllContainedCheckResults() {
        super.clearAllContainedCheckResults();
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.clearAllContainedCheckResults();
        }
    }

    @Override
    public boolean containsError(CheckId checkIdentifier) {
        boolean hasError = super.containsError(checkIdentifier);
        if (hasError) {
            return true;
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            if (bs.containsError(checkIdentifier)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public boolean containsAnyError() {
        boolean hasError = super.containsAnyError();
        if (hasError) {
            return true;
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            if (bs.containsAnyError()) {
                return true;
            }
        }
        return false;
    }

    private void reCreateBoundarySurface(GeometryFactory factory, ParserConfiguration config, BoundarySurface bs) {
        if (bs.getGeometries().isEmpty()) {
            for (AbstractSpaceBoundaryProperty bsp : gmlTunnelElement.getBoundaries()) {
                if (bsp.getObject() != null && bsp.getObject() == bs.getGmlObject()) {
                    logger.warn("Found empty boundary surface: {}, removing from TunnelConstructiveElement", bs.getGmlId());
                    gmlTunnelElement.getBoundaries().remove(bsp);
                    break;
                }
            }
            return;
        }
        bs.reCreateGeometries(factory, config);
    }

    private void setSolidAccordingToLod(Geometry geom, Solid solid) {
        switch (geom.getLod()) {
            case LOD1:
                gmlTunnelElement.setLod1Solid(new SolidProperty(solid));
                break;
            case LOD2:
                gmlTunnelElement.setLod2Solid(new SolidProperty(solid));
                break;
            case LOD3:
                gmlTunnelElement.setLod3Solid(new SolidProperty(solid));
                break;
            default:
                throw new IllegalStateException(CANNOT_ADD + geom.getLod() + " solid to buildings");
        }
    }

    private void setMultiSurfaceAccordingToLod(Geometry geom, MultiSurface ms) {
        switch (geom.getLod()) {
            case LOD0:
                gmlTunnelElement.setLod0MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD2:
                gmlTunnelElement.setLod2MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD3:
                gmlTunnelElement.setLod3MultiSurface(new MultiSurfaceProperty(ms));
                break;
            default:
                throw new IllegalStateException(CANNOT_ADD + geom.getLod() + " multi surface to buildings");
        }
    }

    @Override
    public org.citygml4j.core.model.tunnel.TunnelConstructiveElement getGmlObject() {
        return gmlTunnelElement;
    }

    @Override
    public void unsetGmlGeometries() {
        gmlTunnelElement.setLod0MultiSurface(null);
        gmlTunnelElement.setLod2MultiSurface(null);
        gmlTunnelElement.setLod3MultiSurface(null);
        gmlTunnelElement.setLod1Solid(null);
        gmlTunnelElement.setLod2Solid(null);
        gmlTunnelElement.setLod3Solid(null);
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.unsetGmlGeometries();
        }
    }

    @Override
    public CityObject getTopLevelCityObject() {
        return getParent().getTopLevelCityObject();
    }

    @Override
    public FeatureType getFeatureType() {
        return FeatureType.TUNNEL_CONSTRUCTION_ELEMENT;
    }

    public void addBoundarySurface(BoundarySurface bs) {
        boundarySurfaceList.add(bs);
        bs.setParent(this);
    }

    public List<BoundarySurface> getBoundarySurfaces() {
        return boundarySurfaceList;
    }

    @Override
    public void prepareForChecking() {
        super.prepareForChecking();
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.prepareForChecking();
        }
    }

    @Override
    public void clearMetaInformation() {
        super.clearMetaInformation();
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.clearMetaInformation();
        }
    }

    @Override
    public void collectInstances(CopyHandler handler) {
        super.collectInstances(handler);
        handler.addInstance(boundarySurfaceList);
    }

    @Override
    public void fillValues(Copyable original, CopyHandler handler) {
        super.fillValues(original, handler);
        TunnelConstructiveElement originalTce = (TunnelConstructiveElement) original;
        for (BoundarySurface originalTs : originalTce.boundarySurfaceList) {
            boundarySurfaceList.add(handler.getCopyInstance(originalTs));
        }
    }
}
